/*
 * Copyright (C) 2002, 2003 Jan Panteltje <panteltje@yahoo.com>,
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

// prog for encoding svcd subtitles and multiplexing them into an mpeg
// available under GPL v2

#include "config.h"

#include <assert.h>
#include <ctype.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <netinet/in.h>

#include <sys/stat.h>
#include <sys/types.h>

#include <png.h>


static const char RCSID[]="$Id: //depot/dvdauthor/src/submux.c#23 $";


// (90000*300)/(1260000/2048)
// (9*300)/(126/2048)
// (9*300*2048)/126
#define DVDRATE 43886

#define SUB_BUFFER_MAX		53220 
#define SUB_BUFFER_HEADROOM     1024

#define CVD_SUB_CHANNEL		0x0
#define SVCD_SUB_CHANNEL	0x70
#define DVD_SUB_CHANNEL		0x20

#define DVD_SUB		0
#define CVD_SUB		1
#define SVCD_SUB	2

#define psbufs 10
#define mintime 4

#define maxx 720
#define maxy 576

#ifndef O_BINARY
#define O_BINARY 0 //for mingw
#endif


#define until_next_sub 1 //if 0 subs without length are made when subs are overlapping
#define tbs  90
#define even_xd 1

typedef struct
{
    unsigned char r, g, b, t;
} palt;

typedef struct {
    unsigned char fname[256];
    unsigned int x0, y0, xd, yd; // x0,y0 -- start, xd,yd -- dimension (size)
    int spts, sd, numalpha; // start pts, subtitle duration
    palt pal[4];
} stinfo;

typedef struct
{
    unsigned short int BM;
    unsigned int size;
    unsigned int r0;
    unsigned int ofs;
    unsigned int hsize;
    unsigned int x;
    unsigned int y;
    unsigned short int bplanes;
    unsigned short int bpp;
    unsigned int comp;
    unsigned int imgsize;
    unsigned int xpm,ypm;
    unsigned int cused;
    unsigned int cimp; 
} __attribute__ ((packed)) bmpt;

static unsigned char bb, bc;
static unsigned char *sub, *cbuf;
static unsigned char *img, *fimg;
static FILE *fds;


static unsigned char pptrans[256];
static unsigned int subo;
static unsigned int iline, progr, skip, imgptr, numpal;
static int tofs;
static int debug, svcd;

static int debug_flag=0;

static char header[32];

int dvd_encode(stinfo *s);
int make_stop_packet();
int read_pic(stinfo *s);


void mkpackh(u_int64_t time, unsigned int muxrate, unsigned char stuffing)
{
    unsigned long th=time/300,tl=time%300;
    header[0] = 0x44 | ((th >> 27) & 0x38) | ((th >> 28) & 3);
    header[1] = (th >> 20);
    header[2] = 4 | ((th >> 12) & 0xf8) | ((th >> 13) & 3);
    header[3] = (th >> 5);
    header[4] = 4 | ((th & 31) << 3) | ((tl >> 7) & 3);
    header[5] = 1 | (tl << 1);
    header[6] = muxrate >> 14;
    header[7] = muxrate >> 6;
    header[8] = 3 | (muxrate << 2);
    header[9] = 0xF8 | stuffing;
}


void mkpesh0(unsigned long int pts, unsigned char id)
{   
    header[0] = 0x81;  
    header[1] = 0x80;	//pts flag
    header[2] = 5;
    header[3] = 0x21 | ((pts >> 29) & 6);
    header[4] = pts >> 22;
    header[5] = 1 | (pts >> 14);
    header[6] = (pts >> 7);
    header[7] = 1 | (pts << 1);
    header[8] = id;
}


void mkpesh1(unsigned long int pts, unsigned char id)
{   
    header[0] = 0x81;  
    header[1] = 0x81;	//pts flag + pes extension flag
    header[2] = 8;
    header[3] = 0x21 | ((pts >> 29)&6);
    header[4] = pts >> 22;
    header[5] = 1 | (pts >> 14);
    header[6] = (pts >> 7);
    header[7] = 1 | (pts << 1);
    header[8] = 0x1e; //P - STD_buffer_flag
    header[9] = 0x60; //buffer scale
    header[10] = 0x3a; //buffer size (wtf nu det r..) (svcdverifier tycker  att 64 r ett bra tal hr..)
    header[11] = id;
}

void mkpesh2 (unsigned char id)
{
    header[0] = 0x81;  
    header[1] = 0;
    header[2] = 0;
    header[3] = id;
}


u_int64_t getgts(unsigned char *buf)
{
    u_int64_t th,tl;
    if (((buf[8]&3) != 3) || ((buf[5]&1) != 1) || ((buf[4]&4) != 4)||((buf[2]&4) != 4)|| ((buf[0]&0xc4) != 0x44)) return -1;
    th=(buf[4] >> 3) + (buf[3]*32) + ((buf[2]&3)*32*256) + ((buf[2]&0xf8)*32*128) + (buf[1]*1024*1024) + ((buf[0]&3)*1024*1024*256) + ((buf[0]&0x38)*1024*1024*256*2);
    tl=((buf[4]&3)<<7)|(buf[5]>>1);
    return th*300+tl;
}

void fixgts(u_int64_t *gts,u_int64_t *nextgts)
{
    if( gts[0] < nextgts[0] )
        gts[0]=nextgts[0];
    nextgts[0]=gts[0]+DVDRATE;
}


unsigned int getmuxr(unsigned char *buf)
{
    return (buf[8] >> 2)|(buf[7]*64)|(buf[6]*16384);
}


unsigned int getpts(unsigned char *buf)
{
    if (!(buf[1]&0xc0) || (buf[2]<4) | ((buf[3]&0xe1) != 0x21) || ((buf[5]&1) != 1) || ((buf[7]&1) != 1)) return -1;
    return (buf[7] >> 1) + buf[6]*128 + ((buf[5]&254)*16384) + buf[4]*16384*256 + ((buf[3]&14)*16384*256*128);
}
 
#define nco if (subo<65536) sub[subo++]
#define a2b(b) do { bb = (bb << 2) + b; bc--; if (!bc) { bc = 4;  nco = bb; }} while (0)


int svcd_encode(stinfo *s)
{
    unsigned int x,y,c,l2o;
    double a;
    palt *epal=s->pal;
 
    subo = 2;
    if (s->sd != -1)
    {
	nco = 0x2e;  nco = 0;
	nco = s->sd >> 24; nco = s->sd >> 16;
	nco = s->sd >> 8;  nco = s->sd;
    }
    else
    {
	nco = 0x26;
	nco = 0;
    } 
 
    if (debug > 2)
	fprintf(stderr,\
                "sd: %d   xd: %d  yd: %d  x0: %d  y0: %d\n", s->sd, s->xd, s->yd, s->x0, s->y0); 

    nco = (s->x0 >> 8); nco = s->x0;
    nco = (s->y0 >> 8); nco = s->y0;
    nco = s->xd >> 8; nco = s->xd;
    nco = s->yd >> 8; nco = s->yd;
    for(c = 0;c<4;c++)
    {
	a = (0.257 * (double)epal[c].r) + (0.504*(double)epal[c].g) + (0.098*(double)epal[c].b) + 16;                 
	if (a>255) a = 255; if (a<0) a = 0; 
	nco = a;

	a = (0.439 * (double)epal[c].r) - (0.368 * (double)epal[c].g) - (0.071 * (double)epal[c].b)  +  128; 
	if (a>255) a = 255; if (a<0) a = 0; 
	nco = a;

	a = -(0.148 * (double)epal[c].r) - (0.291 * (double)epal[c].g)  +  (0.439 * (double)epal[c].b) + 128; 
	if (a>255) a = 255; if (a<0) a = 0; 
	nco = a;
  
	nco = epal[c].t;  
    } 
 
    nco = 0; //?????
 
    l2o = subo;
    subo += 2;
    y = 0;
 odd_row: 
    for(; y<s->yd; y += 2)
    {
	bc = 4; bb = 0;
	for(x = 0; x<s->xd;x++)
        {
            if ((c = fimg[y*s->xd+x]) != 0) a2b(c);
            else
            {
                c = 1;
                while (((++x)<s->xd)  &&  (!fimg[y*s->xd+x])) c++;
                x--;
                while (c>4)
                {
                    a2b(0);
                    a2b(3);
                    c -= 4;
                }
                a2b(0);
                a2b(c - 1);
            }
        }
	if (bc != 4) nco = bb << (2*bc);
    }
 
    if (!(y&1))
    {
	if (!(subo&1))
        {
            if (debug>3)
                fprintf(stderr,\
			"padded betweed fields with 1 byte to %d\n",subo%4);
            nco = 0;
        }
	y = 1;
	sub[l2o] = (subo - l2o - 2) >> 8;
	sub[l2o+1] = (subo - l2o - 2);
	goto odd_row;
    } 
 
    nco = 0;// no additional commands
    c = 0;
    while (subo%4)
    {
	nco = 0; c++;
    }
    if (debug>3) fprintf(stderr,"padded with %d byte\n",c);

    sub[0] = subo >> 8;
    sub[1] = subo;
    if (subo == 65536) return -1;
    else return subo;
} /* end function svcd_encode */


#define an(b) do { bb = (bb << 4)+(b); bc--; if (!bc) { bc = 2;  nco = bb; }} while (0)


int cvd_encode(stinfo *s)
{
    unsigned int x, y, c, d;
    double a;
    palt *epal=s->pal;
    int ofs, ofs1=0;
 
    subo = 4;
    ofs = 4;

    c = 0;
    for(y = 0; y < s->yd; y += 2)
    {
    odd_row_cvd: 
	bc = 2; bb = 0;
	for(x = 0; x < s->xd;)
        {
            d = fimg[y * s->xd + x];
            c = 1;
            while (((++x) < s->xd)  &&  (fimg[y * s->xd + x] == d) ) c++;
            if(x == s->xd)
            {
                an(0);
                an(d);
                if (bc != 2) an(0);
                continue;
            }
	   
            while(c > 3)
            {
                an(12 + d);
                c -= 3;
            }
   
            an((c << 2) + d);
        }
    }
 
    if(!(y & 1))
    {
	y = 1;
	ofs1 = subo;
	goto odd_row_cvd;
    }  
 
    sub[2] = subo >> 8;
    sub[3] = subo;

/* setting this to all 0xff then no more subtitles */
    nco = 0x0c;
    nco = 0;
    nco = 0;
    nco = 0;
 
/* set pallette 0-3 */
    for(c = 0; c < 4; c++)
    {
//#define nco if (subo<65536) sub[subo++]
	nco = 0x24 + c;
	if(debug > 3)
        {
            fprintf(stderr, "c=%d R=%.2f G=%.2f B=%.2f\n",\
                    c, (double)epal[c].r, (double)epal[c].g, (double)epal[c].b);
        }

	/* Y */ 
	a = (0.257 * (double)epal[c].r) + (0.504 * (double)epal[c].g) +\
            (0.098 * (double)epal[c].b) + 16.0;
	if (a > 255) a = 255;
	if (a < 0) a = 0;
	nco = a;
	if(debug > 3)
        {
            fprintf(stderr, "Y=%.2f sub[subo]=%ud\n", a, sub[subo - 1]);
        }

	/* U */
	a = (0.439 * (double)epal[c].r) - (0.368 * (double)epal[c].g) -\
            (0.071 * (double)epal[c].b) + 128.0;
	if (a > 255) a = 255;
	if (a < 0) a = 0;
	nco = a;
	if(debug > 3)
        {
            fprintf(stderr, "U=%.2f sub[subo]=%ud\n", a, sub[subo - 1]);
        }

	/* V */
	a = -(0.148 * (double)epal[c].r) - (0.291 * (double)epal[c].g) +\
            (0.439 * (double)epal[c].b) + 128.0;
	if (a > 255) a = 255;
	if (a < 0) a = 0;
	nco = a;
	if(debug > 3)
        {
            fprintf(stderr, "V=%.2f  sub[subo]=%ud (0x%0x)\n",\
                    a, sub[subo - 1], sub[subo - 1]);
        }

    } /* end for pallette 0-3 */
 
/* sethighlight  pallette  */
    for(c = 0; c < 4; c++)
    {
//#define nco if (subo<65536) sub[subo++]
	nco = 0x2c + c;
	if(debug > 3)
        {
            fprintf(stderr, "c=%d R=%.2f G=%.2f B=%.2f\n",\
                    c, (double)epal[c].r, (double)epal[c].g, (double)epal[c].b);
        }

	/* Y */ 
	a = (0.257 * (double)epal[c].r) + (0.504 * (double)epal[c].g) +\
            (0.098 * (double)epal[c].b) + 16;
	if (a > 255) a = 255;
	if (a < 0) a = 0;
	nco = a;
	if(debug > 3)
        {
            fprintf(stderr, "Y=%.2f sub[subo]=%d\n", a, sub[subo - 1]);
        }

	/* U */
	a = (0.439 * (double)epal[c].r) - (0.368 * (double)epal[c].g) -\
            (0.071 * (double)epal[c].b) + 128;
	if (a > 255) a = 255;
	if (a < 0) a = 0;
	nco = a;
	if(debug > 3)
        {
            fprintf(stderr, "U=%.2f sub[subo]=%d\n", a, sub[subo - 1]);
        }

	/* V */
	a = -(0.148 * (double)epal[c].r) - (0.291 * (double)epal[c].g) +\
            (0.439 * (double)epal[c].b) + 128;
	if (a > 255) a = 255;
	if (a < 0) a = 0;
	nco = a;
	if(debug > 3)
        {
            fprintf(stderr, "V=%.2f  sub[subo]=%d\n", a, sub[subo - 1]);
        }

    } /* end for pallette 4-7 */

    if(debug > 3)
    {
	fprintf(stderr,\
                "epal[0].t=%d epal[1].t=%d epal[2].t=%d epal[3].t=%d\n",\
                epal[0].t, epal[1].t, epal[2].t, epal[3].t);
    }

/* x0, y0 */
    nco = 0x17;
    nco = 0xf0 + (s->x0 >> 6);
    nco = (s->x0 << 2) + (s->y0 >> 8);
    nco = s->y0;
 
/* xd, yd */
    nco = 0x1f;
    nco = 0xf0 + ((s->x0 + s->xd - 1) >> 6);
    nco = ((s->x0 + s->xd - 1) << 2) + ((s->y0 + s->yd - 1) >> 8);
    nco = s->y0 + s->yd - 1;
 
/* 0x37 is pallette.t 0-3 contrast */
    nco = 0x37;
    nco = 0xff;
//nco = 0x00;
//nco = 0xff;
    nco = (epal[3].t & 0xf0) + (epal[2].t >> 4);
    nco = (epal[1].t & 0xf0) + (epal[0].t >> 4);

    if(debug > 3)
    {
	fprintf(stderr, "EPALS nco0(2 3h, 2l)=%02x nco1(2 1h,0l)=%02x\n",\
                sub[subo - 1], sub[subo - 2]);
    }

/* 0x3f is high light pallette.t 4-7 contrast */
    nco = 0x3f;
    nco = 0xff;
    nco = 0xff;
    nco = 0xf0;

/* ofs is 4 ? is offset in bitmap to first field data (interlace) */ 
    nco = 0x47;
    nco = 0xff;
    nco = ofs >> 8;
    nco = ofs;
 
/* ofs1 is offset to other field in bitmap (interlace) */
    nco = 0x4f;
    nco = 0xff;
    nco = ofs1 >> 8;
    nco = ofs1;
 
/* unknown!!! (in RX too!) setting all to 0xff keeps the pic on screen! */
    nco = 0x0c;
    nco = 0;
    nco = 0;
    nco = 0; 
 
/* sd, time in display, duration */
    nco = 0x04;
    nco = s->sd >> 16;
    nco = s->sd >> 8;
    nco = s->sd;
 
// IA3
//0: 02 68 02 24
//1: 08 61 08 1d
//2: 08 6b 08 27
//3: 07 6b 07 27
//4: 05 40 04 fc
//5: 08 81 08 3d
//6: 05 4f 15 40


    sub[0] = subo >> 8;
    sub[1] = subo;
    nco = 4;
    nco = 8;
    nco = 12;
    nco = 16;

    if(subo == 65536) return -1;
    else return subo;
} /* end function cvd_encode */

int iswhite(char c)
{
    return c==' '||c=='\t';
}

char *gettok(char **s)
{
    char *r;

    while( iswhite(s[0][0]) )
        s[0]++;
    if( !s[0][0] )
        return 0;
    r=s[0];
    while( !iswhite(s[0][0]) && s[0][0]!=0 )
        s[0]++;
    if( s[0][0] ) {
        s[0][0]=0;
        s[0]++;
    }
    return r;
}

unsigned int parsetime(char *t)
{
    int tf=1;
    int rt=0,n=0,nd=0;

    while(*t) {
        if(isdigit(*t)) {
            if( nd<10000 ) {
                n=n*10+t[0]-'0';
                nd*=10;
            }
        } else if( *t == ':' ) {
            assert(tf);
            rt=rt*60+n;
            n=0;
            nd=1;
        } else if( *t == '.' || *t == ',' ) {
            assert(tf);
            rt=rt*60+n;
            n=0;
            nd=1;
            tf=0;
        }
        t++;
    }
    if( tf )
        return (rt*60+n)*90000;
    else
        return rt*90000+90000*n/nd;
}

void rns(stinfo *n)
{
    char lbuf[256 + 80], *lp,*t;
    int b;

 lns: 
    if (feof(fds) || (!fgets(lbuf, 256 + 80, fds)) )
    {
	n->spts = -1;

	return;
    } 

    iline++;
 
    if (strlen(lbuf) == 256 + 80 - 1) while(getc(fds) != 10);
    if ((lbuf[0] == '#') || (lbuf[0] == ' ') || (lbuf[0] == 0)) goto lns;
    
    lp=lbuf;

    t=gettok(&lp);
    if( !t ) {
	skip++;
	if (debug > -1) fprintf(stderr,"warning, unable to parse line %d, skipping\n", iline);
	goto lns;
    }
    strcpy(n->fname,t);

    t=gettok(&lp);
    if( !t ) {
	skip++;
	if (debug > -1) fprintf(stderr,"warning, unable to parse line %d, skipping\n", iline);
	goto lns;
    }
    n->spts=parsetime(t);

    t=gettok(&lp);
    if( !t ) {
	skip++;
	if (debug > -1) fprintf(stderr,"warning, unable to parse line %d, skipping\n", iline);
	goto lns;
    }
    n->sd=parsetime(t);
 
    if (!n->sd) n->sd = -1;
    else
    {
	if (n->sd <= n->spts)
        {
            if (debug > -1)
            {
                fprintf(stderr, "warning: sub on line %d has end<=start, skipping\n", iline);
            }

            skip++;
            goto lns;
        }
        n->sd -= n->spts;
    }
 
    n->spts += tofs;

    t=gettok(&lp);
    if( !t ) {
	skip++;
	if (debug > -1) fprintf(stderr,"warning, unable to parse line %d, skipping\n", iline);
	goto lns;
    }
    n->x0=atoi(t);

    t=gettok(&lp);
    if( !t ) {
	skip++;
	if (debug > -1) fprintf(stderr,"warning, unable to parse line %d, skipping\n", iline);
	goto lns;
    }
    n->y0=atoi(t);
    
    n->numalpha=0;
    for(b = 0; b < 4; b++)
    {
        t=gettok(&lp);
	if (t) {
            n->pal[b].t = atoi(t);
            n->numalpha++;
        } else
            n->pal[b].t = (b?255:0);

	if (debug > 3)
        {
            fprintf(stderr, "pal: %d %d\n", b, n->pal[b].t);
        }
    }

    if(debug > 2)
    {
	fprintf(stderr, "img: %s, %d (%.2f) %d (%.2f) %d,%d %d,%d\n",
                n->fname, n->spts, (double)n->spts / 90000, n->sd, (double)n->sd / 90000, n->x0, n->y0, n->xd, n->yd);
    }
} /* end function rns */


int imgfix(stinfo *s)
{
    int a, b, i, x0_, y0_, xd_, yd_, x, y, x0fill;
    int cst[4];
    unsigned char c, r, ctrans[4];
    palt epal[4];
 
//xd and yd should be multiples of 4...
    a = 0;
    x0_ = maxx;
    xd_ = 0;
    y0_ = yd_ = -1;
    cst[0] = cst[1] = cst[2] = cst[3] = 0;
  
    for(y = 0; y < s->yd; y++) 
	for(x = 0;x < s->xd; x++)
        {
            if (s->pal[img[a++]].t)
            {
                if (y0_ == -1) y0_ = y;
                if (x < x0_) x0_ = x;
                if (x > xd_) xd_ = x+1;
                yd_ = y+1;
            }
        }
  
    if(yd_ != -1)
    { 
        for(y = y0_; y < yd_; y++) {
            a=y*xd_+x0_;
            for(x =x0_;x < xd_; x++)
                cst[ img[a++] ]++;
        }

	if(svcd) {
            a=0;    
            for( r=1; r<4; r++)
                if( cst[a] < cst[r] )
                    a=r;
            for( i=0; i<4; i++ )
                ctrans[i]=(i+a)&3;
        } else {
            // map the first transparent color to 0
            // map subsequent colors to 2, 3 and 1.
            a=-1;
            for( i=0; i<4; i++ )
                if( !s->pal[i].t && (a==-1 || cst[a]<cst[i]) )
                    a = i;
            if( a==-1 ) {
                a=0;
                for( i=1; i<4; i++ )
                    if( cst[a] < cst[i] )
                        a=i;
            }
            ctrans[a]=0;
            cst[a]=-1;

            for( i=1; i<4; i++ ) {
                a=0;
                for( r=1; r<4; r++ )
                    if( cst[a] < cst[r] )
                        a=r;
                ctrans[a]=i;
                cst[a]=-1;
            }
        }
  
//	if (debug>2)
        // fprintf(stderr,"colors rotated %d steps\n", r);

	for(c = 0; c < 4; c++)
            epal[ctrans[c]] = s->pal[c];
        memcpy(s->pal,epal,sizeof(palt)*4);

	xd_ -= x0_;
	yd_ -= y0_;
  
	b = 0; // fimg ptr
	c = 0; // transparent color
	for(a = 3; a > 0; a--) if (!epal[a].t) c = a;
  
        x0fill=0;
        if( (s->x0 + x0_) & 1 ) {
            if( x0_ )
                x0_--;
            else
                x0fill=1;
        }
  
	if (even_xd && ((xd_+x0fill) & 1))
            xd_++;
  
	if( (s->y0 + y0_) & 1 )
        {
            if (y0_)
                y0_--;
            else {
                memset(fimg+b,c,xd_+x0fill);
                b+=xd_+x0fill;
                yd_++;
                s->y0--;
            } 
        }

	for(y=y0_;y < y0_+yd_; y++)
        {
            a=y*s->xd+x0_;
            if( x0fill ) fimg[b++]=c;
            for(x = 0; x < xd_; x++)
                fimg[b++] = ctrans[img[a++]];
        }

	if (yd_ & 1)
        {
            memset(fimg+b,c,xd_+x0fill);
            b+=xd_+x0fill;
            yd_++;
        }

	s->x0 += x0_;
	s->y0 += y0_;
	s->xd = xd_+x0fill;
	s->yd = yd_;
 
	if (debug > 2)
        {
            fprintf(stderr,
                    "x0: %d  y0: %d  w: %d  h: %d  start: %fs  len: %fs\n",
                    s->x0, s->y0, s->xd, s->yd, (double)s->spts / 90000, (double)s->sd / 90000); 
        }

	return 1;
    } 
 
    return 0;
} /* end function imgfix */

void usage()
{
    fprintf(stderr, "syntax: submux [options] script.sub < in.mpg > out.mpg\n");
    fprintf(stderr, "\t-m <mode>   dvd, cvd, or svcd (only the first letter is checked).\n\t\tDefault is DVD.\n");  	
    fprintf(stderr, "\t-s <stream> number of the substream to insert (default 0)\n");   
    fprintf(stderr, "\t-v <level>  verbosity level (default 0) \n");  
    fprintf(stderr, "\t-P          enable progress indicator\n");
    fprintf(stderr,"\n\tScript file format:\n"
            "\tIMAGE TS_START TS_END X0 Y0 [TRANS_0 TRANS_1 TRANS_2 TRANS_3]\n\n"
            "\tIMAGE - subtitle picture in either PNG with or without alpha, or\n\t\tuncompressed 4 or 8bpp BMP.\n"
            "\tTS_START, TS_END - start and end timestamp in [[hh:]mm:]ss[.ff] format.\n\t\tIf TS_END is 00:00:00.00 then the subtitle never ends.\n"
            "\tX0, Y0 - amount to offset the picture by\n"
            "\tTRANS_0 through 3 - transparency setting, 0=transparent, 255=opaque;\n\t\tdefault is 0 for 0 and 255 for 1-3\n");
    exit(-1);
}

int main(int argc,char **argv)
{
    int fdi, fdo;
    u_int64_t gts, newgts, nextgts;
    unsigned int q, c, a, muxrate, secsize, frame, lps;
    unsigned short int b, vss, subno;
    unsigned char seq, psbuf[psbufs], substr, ncnt;
    int sub_size, optch;
    int bytes_send;
    unsigned short int header_size;
    int mode;
    int max_sub_size;
    stinfo newsti,cursti;
    int lvpts;

    mode = 0; /* default DVD */

    sub = malloc( SUB_BUFFER_MAX + SUB_BUFFER_HEADROOM );
    if(! sub)
    {
	fprintf(stderr, "submux-dvd(): could not allocate space for sub, aborting.\n");

	exit(1);
    }
//fprintf(stderr, "malloc sub=%p\n", sub);

    cbuf = malloc(65536);
    img = malloc(maxx * maxy);
    fimg = malloc(maxx * maxy);
 

    fprintf(stderr,"%s, version %s.  Send bugs to <%s>\n",PACKAGE_NAME,PACKAGE_VERSION,PACKAGE_BUGREPORT);

    gts = 0;
    nextgts = 0;
    tofs = 0;
    progr = 0;
    debug = 0;
    ncnt = 0;
    substr = 0;

    while( -1 != (optch=getopt(argc,argv,"hm:s:v:P")) ) {
        switch( optch ) {
	case 'm':
            switch(optarg[0]) {
            case 'd':
            case 'D':
                mode=DVD_SUB;
                break;

            case 's':
            case 'S':
                mode=SVCD_SUB;
                break;

            case 'c':
            case 'C':
                mode=CVD_SUB;
                break;

            default:
                fprintf(stderr,"mode must be one of dvd, svcd, or cvd\n");
                usage();
            }
            break;

	case 's': substr = atoi(optarg); break;
	case 'v': debug  = atoi(optarg); break;
	case 'P': progr  = 1;            break;

        case 'h': usage();

	default:
            fprintf(stderr,"getopt returned %d\n",optch);
            usage();
        } 
    } /* end switch argv */
 
    switch(mode)
    {
    case DVD_SUB:
    default:	
        svcd = 0;
        substr += DVD_SUB_CHANNEL;
        muxrate = 4400; // 0x1131; // 4400 kbps
        secsize = 2048; //2324;
        break;
    case CVD_SUB:
        svcd = 0;
        substr += CVD_SUB_CHANNEL;
        muxrate = 2600; //0x0a28; // 2600 kbps
        secsize = 2324;
        break;
    case SVCD_SUB:
        svcd = 4;
        substr += SVCD_SUB_CHANNEL;
        muxrate = 4400; //0x1131; // 4400 kbps
        secsize = 2324;
        break;
    } /* end switch mode */

    if( argc-optind!=1 ) {
        fprintf(stderr,"only one argument expected\n");
        usage();
    }

    fdi = 0;
    fds = fopen(argv[optind], "r");
    if(!fds)
    {
	fprintf(stderr, "unable to open file %s\n",  argv[optind]);
	return -1;
    }

    fdo = 1;
 
    if(tofs  &&  (debug > 0) )
	fprintf(stderr, "subtitles offset by %fs\n", (double)tofs / 90000);
 
    iline=0;

    skip=0;
 
    rns(&newsti);

    max_sub_size = 0;
    header_size = 12;
    vss = 0;
    frame = 0;
    lps = 0;
    lvpts = -1;
    subno = -1;
    while(read(fdi, &c, 4) == 4)
    {
	write(fdo, &c, 4);
        c=ntohl(c);
   
	if(progr)
        {
            int i = lseek(fdo, 0, SEEK_CUR);
            if (i > lps + 1024 * 1024 * 10)
                fprintf(stderr, "%d bytes of data written\r", lps = i);
        } 
   
	if(c == 0x1ba) /* packet start code */
        { 
        l_01ba:  
            if(debug > 5) fprintf(stderr, "pack_start_code\n");
            if(read( fdi, psbuf, psbufs) < 1) break;
            newgts = getgts(psbuf);
            if(newgts != -1)
            {
                gts = newgts;
                fixgts(&gts,&nextgts);
                muxrate = getmuxr(psbuf);
            }
            else if (debug >- 1) fprintf(stderr, "incorrect pack header\n");
		
        try_next:   
            /* wait for correct time to insert sub, leave time for vpts to occur */
            if( (lvpts != -1) && (newsti.spts != -1) && (lvpts + 1 * 90000 > newsti.spts) )
		/* 180 ms + 100kb = 200ms -> 380 ms = .38 * 90000 = 34200 */ 
            {
                cursti=newsti;

                /* read bitmap data into img */
                a = read_pic(&cursti);

                if(debug_flag)
                {
                    fprintf(stderr, "after read_bmp(): xd=%d yd=%d x0=%d y0=%d\n", cursti.xd, cursti.yd, cursti.x0, cursti.y0);
                }

                rns(&newsti);

                if( feof(fds) )
                {
                    fprintf(stderr, "Found EOF in .sub file.\n");
                }
    
                if (a)
                {
                    if(debug > -1)
                    {
                        fprintf(stderr, "bad image,  skipping line %d\n", iline - 1);
                    }
	
                    skip++;
                    goto try_next;
                }

                if(cursti.spts < lvpts)
                {
                    if(debug > -1)
                    {
                        fprintf(stderr, \
                                "subtitle time spts=%.2fs < time of first video frame lvpts=%.2fs,  skipping line %d\n",\
                                (double)cursti.spts / 90000.0, (double)lvpts / 90000.0, iline - 1);
                    }

                    skip++;
                    goto try_next;
                }

                if( (cursti.spts < mintime * 90000) && (debug > -1) )
                {
                    fprintf(stderr, \
                            "warning, sub on line %d starts before %ds, (some players won't show it)\n", \
                            iline - 1, mintime);
                }    

                if( !imgfix(&cursti) ) /* data in img to fimg */
                {
                    if (debug > -1)
                    {
                        fprintf(stderr, "blank image, skipping line %d\n", iline - 1);	
                    }

                    skip++;
                    goto try_next;
                }

                if(newsti.spts != -1) /* not last sub */
                {
                    if(cursti.spts + cursti.sd + tbs > newsti.spts)
                    {
                        if (debug > 4)
                            fprintf(stderr, "overlapping sub\n");
                        cursti.sd = -1;
                    }
                } /* end if ! last sub */

                if(debug > 4)
                {
                    fprintf(stderr, "spts: %d  sd: %d  nspts: %d\n",\
                            cursti.spts / 90000, cursti.sd / 90000, newsti.spts / 90000);
                }

                if( (cursti.sd == -1) && (newsti.spts != -1 ) && ( (!svcd) || until_next_sub) )
                {
                    if(newsti.spts > cursti.spts + tbs) cursti.sd = newsti.spts - cursti.spts - tbs;
                    else
                    { 
                        if (debug > -1)
                        {
                            fprintf(stderr,\
                                    "sub with too short or negative duration on line %d, skipping\n",\
                                    iline - 1); 
                        }
                        exit(1);
	
                        skip++;
                        goto try_next; 
                    }
                }

                switch(mode)
                {
                case DVD_SUB:
                    /* rle here */
                    a = dvd_encode(&cursti);
                    break;
                case CVD_SUB:
                    a = cvd_encode(&cursti);
                    break;
                case SVCD_SUB:
                    a = svcd_encode(&cursti);
                    break;
                } /* end switch */

                if(a == -1)
                {
                    if(debug > -1)
                    {
                        fprintf(stderr, "image too large (encoded size>64k), skipping line %d\n", iline - 1);
                    }

                    skip++;
                    goto try_next;
                }

                sub_size = a;     

                if(sub_size > max_sub_size)
                {
                    max_sub_size = sub_size;
                    fprintf(stderr, "max_sub_size=%d\n", max_sub_size);
                }

                seq = 0;
                subno++;

                // header_size is 12 before while starts    

                /* search packet start code */
                bytes_send = 0;
                while(bytes_send != sub_size)
                {
                    int i;

                    /* if not first time here */
                    if(bytes_send)
                    {
                        /* send packet start code */
                        header_size = 4; 
                        c = htonl(0x1ba);
                        write(fdo, &c, 4); // ?
                    }
                    else if (header_size != 12) header_size = 9; // not first time
      

                    /* calculate how many bytes to send */
                    if( (sub_size - bytes_send) > (secsize - 26 - header_size - svcd) ) {
                        if( (sub_size - bytes_send) <= (secsize - 20 - header_size - svcd) )
                            i = secsize - 26 - header_size - svcd;
                        else i = secsize - 20 - header_size - svcd;
                    } else i = sub_size - bytes_send;
	        
                    /* write header */
                    mkpackh(gts, muxrate, 0);
                    fixgts(&gts,&nextgts);

/*
  fprintf(stderr, "system time: %d 0x%lx %d\n", gts, ftell(fds), frame);
  fprintf(stderr, "spts=%d\n", spts);
*/

                    write(fdo, header, 10); 

                    /* write private stream code */
                    c = htonl(0x1bd);
                    write(fdo, &c, 4);     

                    /* write packet length */				
                    b = ntohs(i+header_size+svcd);
                    write(fdo, &b, 2);

                    /* i has NOT changed here! and is still bytes to send */
    
                    if (header_size == 9)
                        mkpesh0(cursti.spts, substr);
                    else if (header_size == 12)
                        mkpesh1(cursti.spts, substr);
                    else
                        mkpesh2(substr);

                    write(fdo, header, header_size); 

                    if(svcd)
                    {
                        /* 4 byte svcd header */
                        unsigned char cc = subno >> 8;

                        write(fdo, &substr, 1); // current subtitle stream

                        if (bytes_send + i == sub_size) seq |= 128;
                        write(fdo, &seq, 1); // packet number in current sub
                        // 0 - up, last packet has bit 7 set
                        write(fdo, &cc, 1);    // h current sub nr 
                        write(fdo, &subno, 1); // l current sub nr
                    } 

                    seq++;

                    /* write i data bytes, increment bytes_send by bytes written */
                    bytes_send += write(fdo, sub + bytes_send, i);

                    /* test if full sector */
                    if (i != secsize - 20 - header_size - svcd)
                    {
                        unsigned short bs;
                        /* if sector not full, write padding? */

                        /* write padding code */
                        c = htonl(0x1be);
                        write(fdo, &c, 4);     

                        /* calculate number of padding bytes */
                        b = secsize - i - header_size - 26 - svcd;

                        if(debug > 4)
                        {
                            fprintf(stderr, "padding, b: %d\n", b);
                        }
					
                        /* write padding stream size */
                        bs = htons(b);			//fixa
                        write(fdo, &bs, 2);

                        /* write padding end marker ? */
                        c = 0xff;
                        for(q = 0; q < b; q++) write(fdo, &c, 1);
                    } 
                } /* end while bytes_send ! sub_size */
     
                /* write packet start code */
                c = htonl(0x1ba);
                write(fdo, &c, 4);        

                if (debug > 0)
                {
                    fprintf(stderr,"subtitle inserted at: %f sd=%d\n", (double)cursti.spts / 90000, cursti.sd / 90000);
                }

            } /* end if valid sub */
		 
            if (debug > 5)
            {
                fprintf(stderr, "system time: %lld 0x%lx %d\n",\
			gts, ftell(fds), frame);
            }
            mkpackh(gts,muxrate,0);
            write(fdo, header, psbufs);
        }
	else if( c>=0x1bb && c<=0x1ef )
        {
            if (read(fdi, &b, 2) != 2) break;
            write(fdo, &b, 2);
            b = ntohs(b);
            
            if (read(fdi, cbuf, b) != b) break;
            write(fdo, cbuf, b);

            if( c == 0x1e0 ) {
                if (debug > 5) fprintf(stderr, "video stream\n"); 
                a = getpts(cbuf);
                if (a != -1) {
                    if( lvpts == -1 ) {
                        newsti.spts+=a;
                        tofs=a;
                    }
                    lvpts = a;
                }
            }
        } else {
            if (debug > 0)
            {
                fprintf(stderr, "unknown header %.2x %.2x %.2x %.2x\n",\
                        c & 255, (c >> 8) & 255, (c >> 16) & 255, (c >> 24) & 255);   
            }

            a = b = 0;
            while(a  !=  0x1ba)
            {
                unsigned char nc;

                if (read(fdi, &nc, 1) < 1) break;

                write(fdo, &nc, 1);

                a=(a<<8)|nc;

                if(debug > 6) fprintf(stderr, "0x%x\n", a);
                b++;
            }
            fprintf(stderr, "skipped %d bytes of garbage\n", b);

            goto l_01ba;
        }

    } /* end while read / write all */
 
    fprintf(stderr, "max_sub_size=%d\n", max_sub_size);

    if (subno  !=  0xffff)
    {
	fprintf(stderr,\
                "%d subtitles added, %d subtitles skipped, stream: %d, offset: %.2f\n",\
                subno + 1, skip, substr, (double)tofs / 90000);
    }
    else
    {
	fprintf(stderr, "warning: no subtitles added\n"); 
    }
      
    fclose(fds);

    return 0;
} /* end function main */

void initpp()
{
    int i;

    for( i=0; i<256; i++) 
        pptrans[i]=255;
    numpal=0;
    imgptr=0;
}

void setrow(stinfo *s,int y)
{
    imgptr=y*s->xd;
}

void putpixel(int c)
{
    if( pptrans[c]==255 ) {
        if( numpal==4 )
            pptrans[c]=4;
        else
            pptrans[c]=numpal++;
    }
    img[imgptr++]=pptrans[c]&3;
}

void putpal(stinfo *s,int c,palt *p)
{
    if( pptrans[c]<4 ) {
        palt *d=&s->pal[pptrans[c]];
        d->r=p->r;
        d->g=p->g;
        d->b=p->b;
    }
}

int read_png(stinfo *s)
{
    unsigned char pnghead[8];
    FILE *fp;
    png_struct *ps;
    png_info *pi;
    png_byte **rowp;
    unsigned long width,height;
    int bit_depth,color_type,numpal,channels,x,y,i;
    palt pal[4];

    fp=fopen(s->fname,"r");
    if( !fp ) {
	fprintf(stderr,"error, unable to open file %s\n",s->fname);
	return -1;
    }
    fread(pnghead,1,8,fp);
    if(png_sig_cmp(pnghead,0,8)) {
	fprintf(stderr,"error, file %s isn't a png\n",s->fname);
	fclose(fp);
	return -1;
    }
    ps=png_create_read_struct(PNG_LIBPNG_VER_STRING,NULL,NULL,NULL);
    if( !ps ) {
        fprintf(stderr,"error initializing png\n");
        fclose(fp);
        return -1;
    }
    pi=png_create_info_struct(ps);
    if( !pi ) {
        fprintf(stderr,"error initializing png\n");
        png_destroy_read_struct(&ps,NULL,NULL);
        fclose(fp);
        return -1;
    }
    png_init_io(ps,fp);
    png_set_sig_bytes(ps,8);

    png_read_png(ps,pi,PNG_TRANSFORM_PACKING|PNG_TRANSFORM_STRIP_16|PNG_TRANSFORM_EXPAND,NULL);
    rowp=png_get_rows(ps,pi);
    png_get_IHDR(ps,pi,&width,&height,&bit_depth,&color_type,NULL,NULL,NULL);
    // format is now RGB[A] or G[A]
    channels=png_get_channels(ps,pi);
    fclose(fp);
    if(color_type&PNG_COLOR_MASK_COLOR)
        channels-=3;
    else
        channels--;
    if(color_type&PNG_COLOR_MASK_ALPHA)
        channels--;
    assert(bit_depth==8); // 8bpp, not 1, 2, 4, or 16
    assert(!(color_type&PNG_COLOR_MASK_PALETTE)); // not a palette
    fprintf(stderr,"w=%lu, h=%lu, bd=%d, ct=%d\n",width,height,bit_depth,color_type);
    if( width>maxx || height>maxy ) {
        fprintf(stderr,"png %s is too big: %lux%lu\n",s->fname,width,height);
        png_destroy_read_struct(&ps,&pi,NULL);
        return -1;
    }
    s->xd=width;
    s->yd=height;
    numpal=0;
    memset(pal,0xaa,sizeof(palt)*4);
    for( y=0; y<height; y++ ) {
        unsigned char *d=rowp[y];
        for( x=0; x<width; x++ ) {
            palt p;

            if(color_type&PNG_COLOR_MASK_COLOR) {
                p.r=*d++;
                p.g=*d++;
                p.b=*d++;
            } else {
                p.r=*d++;
                p.g=p.r;
                p.b=p.r;
            }
            if( color_type&PNG_COLOR_MASK_ALPHA )
                p.t=*d++;
            else
                p.t=255;
            d+=channels;
            for( i=0; i<numpal; i++ )
                if( !memcmp(&pal[i],&p,sizeof(palt)) )
                    break;
            if( i==numpal ) {
                if( numpal==4 ) {
                    fprintf(stderr,"png %s exceeded 4 colors\n",s->fname);
                    png_destroy_read_struct(&ps,&pi,NULL);
                    return -1;
                }
                pal[numpal++]=p;
            }
            img[y*width+x]=i;
        }
        // free(rowp[y]);
    }
    // free(rowp);
    if(!(color_type&PNG_COLOR_MASK_ALPHA))
        i=4;
    else
        i=s->numalpha;
    while((--i)>=0)
        pal[i].t=s->pal[i].t;
    memcpy(s->pal,pal,4*sizeof(palt));
    png_destroy_read_struct(&ps,&pi,NULL);
    fprintf(stderr,"png had %d colors\n",numpal);

    return 0;
}

int read_bmp(stinfo *s)
{
    bmpt bmp;
    int fdr;
    unsigned int d, i, x, y, size, pals;
    unsigned short int b;
    unsigned char c;
    unsigned char *limg;
    palt lpal[256];
    int pallette_start;

    fdr = open(s->fname, O_RDONLY | O_BINARY);
    if(fdr < 0)
    {
	fprintf(stderr,"error, unable to open file %s\n",s->fname);
	return -1;
    } 
 
    if(read(fdr, &bmp, 54) < 54)
    {
	fprintf(stderr,"error, file too short: %s\n",s->fname);
	close(fdr);
	return -1;
    }	
 
    if(bmp.BM  !=  19778)
    {
	fprintf(stderr,"error, file %s isn't a bmp\n",s->fname);
	close(fdr);
	return -1;
    }
 
    if(debug > 3) fprintf(stderr,"xd: %d  yd: %d\n", bmp.x, bmp.y);
 
    s->xd = bmp.x;
    s->yd = bmp.y;

    if( (bmp.x > maxx) || (bmp.y > maxy) )
    {
	fprintf(stderr,"error, image larger than 720x576\n");
	close(fdr);
	return -1;
    } 

    if( (bmp.bpp  !=  4)  &&  (bmp.bpp  !=  8) )
    {
	fprintf(stderr,"error, only 4 & 8bpp images supported\n");
	close(fdr);
	return -1;
    }

    if(bmp.comp)
    {
	fprintf(stderr,"error, only uncompressed images are supported\n");
	close(fdr);
	return -1;
    }

    d = 0;
 
//lseek(fdr, bmp.hsize - 40, SEEK_CUR);
 
/*
  start color table
  file header size + info header size
*/

    pallette_start = 14 + bmp.hsize; // hsize = BitmapInfoHeader size!

    if(debug > 3)
    {
	fprintf(stderr, "WAS name=%s pallette_start=%d\n", s->fname, pallette_start);
    }

    lseek(fdr, pallette_start, SEEK_SET);

    pals = 1  <<  bmp.bpp;

    for(b = 0; b < pals; b++)
    {
	d += read(fdr, &lpal[b].b, 1); 
	d += read(fdr, &lpal[b].g, 1); 
	d += read(fdr, &lpal[b].r, 1); 
	d += read(fdr, &c, 1);
	if(debug > 3)
        {
            fprintf(stderr, "WAS lpal[b].b=%d lpal[b].g=%d lpal[b].r=%d c=%d d=%d\n",\
                    lpal[b].b, lpal[b].g, lpal[b].r, c, d);
        }
    }
 
    if(d != pals * 4)
    {
	fprintf(stderr,"error file too short: %s\n",s->fname);
	close(fdr);
	return -1;
    }	
 
    initpp();

    if(bmp.bpp  ==  4)
    {
	size = ((s->xd+7)&(-8)) / 2;

	limg = malloc(size);
  
	lseek(fdr, bmp.ofs, SEEK_SET);
  
	for(y = s->yd - 1; y  !=  -1; y--)
        {
            if( read(fdr, limg, size)  !=  size)
            {
                fprintf(stderr,"error, file too short: %s\n",s->fname);
                free(limg);
                close(fdr);
                return -1;
            }

            setrow(s,y);
            for(x = 0; x < s->xd; x += 2)
            {
                putpixel(limg[x>>1] >> 4);
                if(x + 1  !=  s->xd)
                    putpixel(limg[x>>1]&15);
            }
        } 
    } /* end if 4 bpp */
    else /* 8 bits per pixel */
    {
 	size = (s->xd+3)&(-4);

	limg = malloc(size);
  
	lseek(fdr, bmp.ofs, SEEK_SET);
  
	for(y = s->yd - 1; y != -1; y--)
        {
            setrow(s,y);
            if(read(fdr, limg, size)  !=  size)
            {
                fprintf(stderr,"error, file too short: %s\n",s->fname);
                free (limg);
                close(fdr);
                return -1;
            }
            for(x = 0; x < s->xd; x++)
                putpixel(limg[x]);
        } 
    } /* end if 8 bpp */
 
    free(limg);
    close(fdr); 

    if(debug > 3)
    {
	for(i = 0; i < pals; i++)
        {
            fprintf(stderr,\
                    "WAS 1 lpal[%d].b=%d lpal[%d].g=%d lpal[%d].r=%d c=%d d=%d\n",\
                    i, lpal[i].b, i, lpal[i].g, i, lpal[i].r, c, d);
        }
    }
    for( i=0; i<pals; i++ )
        putpal(s,i,&lpal[i]);

    return 0;
} /* end function read_bmp */

int read_pic(stinfo *s)
{
    int l=strlen(s->fname);
    if( l>=4 ) {
        if(!strcasecmp(s->fname+l-4,".png"))
            return read_png(s);
        if(!strcasecmp(s->fname+l-4,".bmp"))
            return read_bmp(s);
    }
    // default is png?
    return read_png(s);
}


static unsigned char *acptr;
static int oddp=0;
void store_nibble(int val)
{
    assert(val>=0 && val<=15);

    if(! oddp)
	*acptr = (val << 4);
    else
	*acptr++ |= val;

    oddp = 1 - oddp;
} /* end function store_nibble */


void store_trinibble(int val)
{
    store_nibble((val>>8)&15);
    store_nibble((val>>4)&15);
    store_nibble(val&15);
}

void store_2(int val)
{
    *acptr++=val>>8;
    *acptr++=val;
}


void do_rle(int count, int color)
{
    int a;

    /* argument check */
    assert(count>=1 && count<=255);
    assert(color>=0 && color<=3);

    /* make rle code in b */
    a = (count << 2) | color;

    /* a now ranges from 0x4 up, because count is at least 1 */

    if(count >= 64)		// 64 - 255
    {
	/* 64-255, 16 bits, 0 0 0 0  0 0 n n  n n n n  n n c c */
	store_nibble(0);
	store_nibble( (a & 0xf00) >> 8);
	store_nibble( (a & 0xf0) >> 4);
	store_nibble( a & 0xf);
    }
    else if(count >= 16)	// 16 - 63
    {
	/* 16 - 63, 12 bits, 0 0 0 0  n n n n  n n c c */
	store_nibble( 0);
	store_nibble( (a & 0xf0) >> 4);
	store_nibble( a & 0xf);
    }
    else if(count >= 4)		// 4 - 15
    {
	/* 4-15, 8 bits, 0 0 n n  n n c c */
	store_nibble( (a & 0xf0) >> 4);
	store_nibble( a & 0xf);
    }
    else			// 1 - 3
    {
	/* 1-3, 4 bits, n n c c */
	store_nibble( a & 0xf );
    }
} /* end function do_rle */


void dvd_encode_row(int y,int xd,unsigned char *icptr)
{
    int new_pos, x;

    new_pos = 0;
    icptr+=y*xd;
    for(x = 0; x < xd-1; x++)
    {
        /* get color from x,y */

        if(icptr[x + 1] != icptr[x])
        {
            int count=x+1-new_pos;
            while(count>255) {
                do_rle(255,icptr[x]);
                count-=255;
            }
            if( count )
                do_rle(count,icptr[x]);
            new_pos = x + 1;
        }
    } /* end for x */

    /*
      One special case,
      encoding a count of zero using the 16-bit format,
      indicates the same pixel value until the end of the line. 
    */
	
    if( xd != new_pos )
    {
        int count=xd-new_pos;

        if( count < 64 )
            do_rle(count,icptr[new_pos]);
        else {
            /* send same colors to end of line */
            store_nibble(0);
            store_nibble(0);
            store_nibble(0);
            store_nibble(icptr[new_pos]);
        }
    }

    /*
      If, at the end of a line, the bit count is not a multiple of 8, four fill bits of 0 are added.
    */

    if(oddp)
        store_nibble(0);
    if( acptr-sub >= SUB_BUFFER_MAX ) {
        fprintf(stderr,"image is too big for the subtitle buffer\n");
        exit(1);
    }
}

int dvd_encode(stinfo *s)
{
    int a;
    int subtitle_packet_size;
    int xstart, xsize;
    int ystart, ysize;
    unsigned char *next_command_ptr;
    int offset0, offset1;
    int pattern, background, emphasis1, emphasis2;
    int y;
    unsigned char *icptr;

    xstart = s->x0;
    xsize = s->xd;
    ystart = s->y0;
    ysize = s->yd;


/*
  720 x 576 = 414720 bytes, for a 2 bit bitmap = 103680 bytes, compressing MUST reduce this to less then 65536 bytes,
  or a 2 byte number - overhead 65535 - 65507 = 28 bytes, as max. 65507 is OK).
  This 23 bytes control buffer, 6 bytes end sequence?
*/

    /* encode the .ppm to DVD run length encoded format */
    /* for all bytes in img */

/*
  (X >> 2) is the number of pixels to display, and (X & 0x3)
  is the color of the pixel.
*/

    icptr = fimg;  
//icptr = img;// use if call to imgfix() commented out

    acptr = sub;
    //2 bytes packet size, to be filled in later
    //2 bytes pointer to control area, to be filled in later
    acptr += 4;

    /* copy image data to sub */
    offset0=acptr-sub;
    for( y=0; y<s->yd; y+=2 )
        dvd_encode_row(y,s->xd,icptr);

    offset1=acptr-sub;
    for( y=1; y<s->yd; y+=2 )
        dvd_encode_row(y,s->xd,icptr);

    /* start first command block */
/* 
   control area starts here, with same pointer
   SP_DCSQT
   Sub-Picture Display Control SeQuence Table
   This area contains blocks (SP_DCSQ) of commands to the decoder.
   Each SP_DCSQ begins with a 2 word header

   offset  name            contents
   0       SP_DCSQ_STM     delay to wait before executing these commands.
   The units are 90KHz clock (same as PTM) divided by 1024 - see below.

   2       SP_NXT_DCSQ_SA  offset within the Sub-Picture Unit to the next SP_DCSQ.
   If this is the last SP_DCSQ, it points to itself.

   Converting frames and time to SP_DCSQ_STM values
   The direct method of converting time to delay values is to multiply time in seconds by 90000/1024 and
   truncate the value.
   Rounding up will cause the display to occur one frame late.
*/

    /* set pointer to this command block */
    a = acptr - sub;
    sub[2] = a >> 8;
    sub[3] = a;

    store_2(0); // delay to wait before executing this command block

    /* remember position, will set later */
    next_command_ptr = acptr;
    acptr++; // pointer to next command block, 2 bytes, to be filled in from next command block. 
    acptr++;

    /* command 0, forced start display, 1 byte. */
    if( s->sd<0 )
        *acptr++ = 0; // menu buttons NEED this to be set

    /* command 1, start display, 1 byte */
    if( s->sd>=0 )
        *acptr++ = 1;

    /* selected palettes for pixel value */

    /* DVD definitions pixel values */
    background	= 0;
    pattern	= 1;
    emphasis1	= 2;
    emphasis2	= 3;

    /* command 3, palette, 3 bytes */
    *acptr++ = 3;
    store_nibble(emphasis2);
    store_nibble(emphasis1);
    store_nibble(pattern);
    store_nibble(background);

    /* command 4, alpha blend, contrast / transparency t_palette, 3 bytes  0, 15, 15, 15 */

    *acptr++ = 4;
    store_nibble(s->pal[3].t>>4);
    store_nibble(s->pal[2].t>>4);
    store_nibble(s->pal[1].t>>4);
    store_nibble(s->pal[0].t>>4);

    /* command 5, display area, 7 bytes from: startx, xsize, starty, ysize */

    *acptr++ = 5;
    store_trinibble(xstart);
    store_trinibble(xstart+xsize-1);
    store_trinibble(ystart);
    store_trinibble(ystart+ysize-1);

    /* command 6, image offsets, 5 bytes */
    *acptr++ = 6;

    store_2(offset0);
    store_2(offset1);

    /* command 0xff, end command block, 1 byte, */
    *acptr++ = 0xff;

    /* end first command block */


    /* start second command block */

    if( s->sd>=0 ) {
        /* set pointer in previous block to point to this position */
        a = acptr - sub;
        next_command_ptr[0] = a >> 8;
        next_command_ptr[1] = a;

        /* update pointer to this command block */
        next_command_ptr = acptr;

        /* delay to wait before executing next comand */
        store_2(s->sd / 1024);

        /* last block, point to self */
        store_2(next_command_ptr - sub);

        /* stop command (executed after above delay) */
        *acptr++ = 0x02;

        /* end command block command */
        *acptr++ = 0xff;
    } else {
        a=next_command_ptr-sub-2;
        next_command_ptr[0] = a >> 8;
        next_command_ptr[1] = a;
    }

    /* end second commmand block */


    /* set the packet size */
    subtitle_packet_size = acptr - sub;

    if( subtitle_packet_size > SUB_BUFFER_MAX ) {
        fprintf(stderr,"image is too big for the subtitle buffer\n");
        exit(1);
    }

//#define MAKE_DVD_SUBTITLE_PACKET_EVEN
#ifdef MAKE_DVD_SUBTITLE_PACKET_EVEN
/* make size even if odd */
    if(subtitle_packet_size & 1)
    {
	if(debug_flag)
        {
            fprintf(stderr, "WAS ODD\n");
        }

	/* only if odd length, to make it even */
	cptr++;
	*acptr = 0xff;

	subtitle_packet_size++;
    }
#endif

    /* set subtitle packet size */
    a = subtitle_packet_size;
    sub[0] = a >> 8;
    sub[1] = a;

    return a;
} /* end function dvd_encode */
